import codecs
import typing
import uuid

class CodecContext:
    def get_text_codec(self) -> codecs.CodecInfo: ...
    def get_json_decoder(self) -> object: ...
    def get_json_encoder(self) -> object: ...

@typing.final
class ReadBuffer: ...

@typing.final
class WriteBuffer: ...

class BufferError(Exception): ...

@typing.final
class UUID(uuid.UUID):
    def __init__(self, inp: typing.AnyStr) -> None: ...
